@@ -1,23 +1,44 @@
 import re
-import structlog
 import time
 
+# graphite-api and graphite-web have different logging systems
 try:
     from graphite_api.intervals import Interval, IntervalSet
     from graphite_api.node import LeafNode, BranchNode
+    import structlog
+    logger = structlog.get_logger()
 except ImportError:
     from graphite.intervals import Interval, IntervalSet
     from graphite.node import LeafNode, BranchNode
+    from graphite.logger import log
+    import logging
+
+    class StructlogCompat(object):
+        def __init__(self):
+            log.debugLogger = logging.getLogger("debug")
+
+        @staticmethod
+        def debug(*args, **kwargs):
+            if not log.infoLogger.isEnabledFor(logging.DEBUG):
+                return
+
+            args_list = []
+            for arg in args:
+                args_list.append(arg)
+            for key, value in kwargs.iteritems():
+                args_list.append(key + '=' + str(value))
+            return log.infoLogger.debug(' '.join(args_list))
+    logger = StructlogCompat()
+# uncomment following to enable debugging logger
+#    log.infoLogger.setLevel(logging.DEBUG)
 
 from influxdb import InfluxDBClient
 
-logger = structlog.get_logger()
-
 # uncomment the following block if you suspect that logger output is often not visible for some reason, but printing works, when you ctrl-c gunicorn at least (this is still a mystery to me)
 
 # def debug(*args, **kwargs):
 #    import pprint
-#     pprint.pprint((args, kwargs))
+#    pprint.pprint((args, kwargs))
 # logger.debug = debug
 
 
@@ -50,7 +71,6 @@ except:
 # statsd = StatsClient("host", 8125)
 
 
-
 def print_time(t=None):
     """
     t unix timestamp or None
@@ -60,36 +80,35 @@ def print_time(t=None):
     return "%d (%s)" % (t, time.ctime(t))
 
 
-def config_to_client(config=None):
+def normalize_config(config=None):
+    ret = {}
     if config is not None:
         cfg = config.get('influxdb', {})
-        host = cfg.get('host', 'localhost')
-        port = cfg.get('port', 8086)
-        user = cfg.get('user', 'graphite')
-        passw = cfg.get('pass', 'graphite')
-        db = cfg.get('db', 'graphite')
-        cheat_times = cfg.get('cheat_times', False)
+        ret['host'] = cfg.get('host', 'localhost')
+        ret['port'] = cfg.get('port', 8086)
+        ret['user'] = cfg.get('user', 'graphite')
+        ret['passw'] = cfg.get('pass', 'graphite')
+        ret['db'] = cfg.get('db', 'graphite')
+        ret['schema'] = cfg.get('schema', [])
     else:
         from django.conf import settings
-        host = getattr(settings, 'INFLUXDB_HOST', 'localhost')
-        port = getattr(settings, 'INFLUXDB_PORT', 8086)
-        user = getattr(settings, 'INFLUXDB_USER', 'graphite')
-        passw = getattr(settings, 'INFLUXDB_PASS', 'graphite')
-        db = getattr(settings, 'INFLUXDB_DB', 'graphite')
-        cheat_times = getattr(settings, 'INFLUXDB_CHEAT_TIMES', False)
-
-    return InfluxDBClient(host, port, user, passw, db), cheat_times
+        ret['host'] = getattr(settings, 'INFLUXDB_HOST', 'localhost')
+        ret['port'] = getattr(settings, 'INFLUXDB_PORT', 8086)
+        ret['user'] = getattr(settings, 'INFLUXDB_USER', 'graphite')
+        ret['passw'] = getattr(settings, 'INFLUXDB_PASS', 'graphite')
+        ret['db'] = getattr(settings, 'INFLUXDB_DB', 'graphite')
+        ret['schema'] = getattr(settings, 'INFLUXDB_SCHEMA', [])
+    return ret
 
 
 class InfluxdbReader(object):
-    __slots__ = ('client', 'path', 'step', 'cache', 'cheat_times')
+    __slots__ = ('client', 'path', 'step', 'cache')
 
-    def __init__(self, client, path, step, cache, cheat_times):
+    def __init__(self, client, path, step, cache):
         self.client = client
         self.path = path
         self.step = step
         self.cache = cache
-        self.cheat_times = cheat_times
 
     def fetch(self, start_time, end_time):
         # in graphite,
@@ -189,53 +208,9 @@ class InfluxdbReader(object):
         return datapoints
 
     def get_intervals(self):
-        if self.cheat_times:
-            logger.debug(caller='get_intervals', msg="cheat_times enabled. this is gonna be quick..", debug_key=self.path)
             now = int(time.time())
             return IntervalSet([Interval(1, now)])
 
-        key_first = "%s_first" % self.path
-        first = self.cache.get(key_first)
-        if first is None:
-            logger.debug(caller='get_intervals', cache_miss=key_first, debug_key=self.path)
-            q = 'select * from "%s" limit 1 order asc' % self.path
-            with statsd.timer('service=graphite-api.ext_service=influxdb.target_type=gauge.unit=ms.what=query_individual_first_point_duration'):
-                first_data = self.client.query(q)
-            first = 0
-            valid_res = False
-            try:
-                first = first_data[0]['points'][0][0]
-                valid_res = True
-            except Exception:
-                pass
-            if valid_res:
-                logger.debug(caller='get_intervals', cache_add=key_first, debug_key=self.path)
-                self.cache.add(key_first, first, timeout=self.step * 1000)
-        else:
-            logger.debug(caller='get_intervals', cache_hit=key_first, debug_key=self.path)
-
-        key_last = "%s_last" % self.path
-        last = self.cache.get(key_last)
-        if last is None:
-            logger.debug(caller='get_intervals', cache_miss=key_last, debug_key=self.path)
-            q = 'select * from "%s" limit 1' % self.path
-            with statsd.timer('service=graphite-api.ext_service=influxdb.target_type=gauge.unit=ms.what=query_individual_last_point_duration'):
-                last_data = self.client.query(q)
-            last = 0
-            valid_res = False
-            try:
-                last = last_data[0]['points'][0][0]
-                valid_res = True
-            except Exception:
-                pass
-            if valid_res:
-                logger.debug(caller='get_intervals', cache_add=key_last, debug_key=self.path)
-                self.cache.add(key_last, last, timeout=self.step)
-        else:
-            logger.debug(caller='get_intervals', cache_hit=key_last, debug_key=self.path)
-
-        return IntervalSet([Interval(first, last)])
-
 
 class InfluxLeafNode(LeafNode):
     __fetch_multi__ = 'influxdb'
@@ -243,7 +218,7 @@ class InfluxLeafNode(LeafNode):
 
 class InfluxdbFinder(object):
     __fetch_multi__ = 'influxdb'
-    __slots__ = ('client', 'schemas', 'cache', 'cheat_times')
+    __slots__ = ('client', 'schemas', 'cache')
 
     def __init__(self, config=None):
         try:
@@ -253,50 +228,46 @@ class InfluxdbFinder(object):
             from django.core.cache import cache
             self.cache = cache
 
-        self.client, self.cheat_times = config_to_client(config)
-        # we basically need to replicate /etc/carbon/storage-schemas.conf here
-        # so that for a given series name, and given from/to, we know the corresponding steps in influx
-        # for now we assume we don't do continuous queries yet, and have only one resolution per match-string
-        # for now, edit your settings here, TODO make this properly configurable or have it stored in influx and query influx
-        self.schemas = [(re.compile(''), 60)]
+        config = normalize_config(config)
+        self.client = InfluxDBClient(config['host'], config['port'], config['user'], config['passw'], config['db'])
+        self.schemas = [(re.compile(patt), step) for (patt, step) in config['schema']]
 
     def assure_series(self, query):
-
-        regex = self.compile_regex(query)
+        regex = self.compile_regex(query, True)
 
         key_series = "%s_series" % query.pattern
         with statsd.timer('service=graphite-api.action=cache_get_series.target_type=gauge.unit=ms'):
             series = self.cache.get(key_series)
         if series is not None:
-            return series, regex
+            return series
 
-        # if not in cache, generate from scratch :(
+        # if not in cache, generate from scratch
         # first we must load the list with all nodes
-        with statsd.timer('service=graphite-api.action=cache_get_nodes.target_type=gauge.unit=ms'):
-            nodes = self.cache.get("influxdb_list_series")
-            if nodes is None:
-                raise Exception("series not in cache. please run maintain_cache.py")
-
-        # and then build the sublist of all matching ones
-        with statsd.timer('service=graphite-api.action=find_series.target_type=gauge.unit=ms'):
-            series = [name for name in nodes if regex.match(name) is not None]
+        with statsd.timer('service=graphite-api.ext_service=influxdb.target_type=gauge.unit=ms.action=get_series'):
+            ret = self.client.query("list series /%s/" % regex.pattern)
+            # as long as influxdb doesn't have good safeguards against series with bad data in the metric names, we must filter out like so:
+            series = [serie[1] for serie in ret[0]['points'] if serie[1].encode('ascii', 'ignore') == serie[1]]
 
         # store in cache
         with statsd.timer('service=graphite-api.action=cache_set_series.target_type=gauge.unit=ms'):
             self.cache.add(key_series, series, timeout=300)
+        return series
 
-        return series, regex
+    def compile_regex(self, query, series=False):
+        # we turn graphite's custom glob-like thing into a regex, like so:
+        # * becomes [^\.]*
+        # . becomes \.
 
-    def compile_regex(self, query):
-        # query.pattern is basically regex, though * should become [^\.]*
-        # and . \.
-        # but list series doesn't support pattern matching/regex yet
-        regex = '^{0}$'.format(
-            query.pattern.replace('.', '\.').replace('*', '[^\.]*')
+        if series:
+            regex = '^{0}'
+        else:
+            regex = '^{0}$'
+
+        regex = regex.format(
+            query.pattern.replace('.', '\.').replace('*', '[^\.]*').replace('{', '(').replace(',', '|').replace('}', ')')
         )
         logger.debug("searching for nodes", pattern=query.pattern, regex=regex)
-        regex = re.compile(regex)
-        return regex
+        return re.compile(regex)
 
     def get_leaves(self, query):
         key_leaves = "%s_leaves" % query.pattern
@@ -304,18 +275,20 @@ class InfluxdbFinder(object):
             data = self.cache.get(key_leaves)
         if data is not None:
             return data
-        series, regex = self.assure_series(query)
+        series = self.assure_series(query)
+        regex = self.compile_regex(query)
         logger.debug(caller="get_leaves", key=key_leaves)
         leaves = []
         with statsd.timer('service=graphite-api.action=find_leaves.target_type=gauge.unit=ms'):
             for name in series:
-                logger.debug("found leaf", name=name)
-                res = 10
-                for (rule_patt, rule_res) in self.schemas:
-                    if rule_patt.match(name):
-                        res = rule_res
-                        break
-                leaves.append([name, res])
+                if regex.match(name) is not None:
+                    logger.debug("found leaf", name=name)
+                    res = 60  # fallback default
+                    for (rule_patt, rule_res) in self.schemas:
+                        if rule_patt.match(name):
+                            res = rule_res
+                            break
+                    leaves.append([name, res])
         with statsd.timer('service=graphite-api.action=cache_set_leaves.target_type=gauge.unit=ms'):
             self.cache.add(key_leaves, leaves, timeout=300)
         return leaves
@@ -327,7 +300,8 @@ class InfluxdbFinder(object):
             data = self.cache.get(key_branches)
         if data is not None:
             return data
-        series, regex = self.assure_series(query)
+        series = self.assure_series(query)
+        regex = self.compile_regex(query)
         logger.debug(caller="get_branches", key=key_branches)
         branches = []
         with statsd.timer('service=graphite-api.action=find_branches.target_type=gauge.unit=ms'):
@@ -344,15 +318,19 @@ class InfluxdbFinder(object):
         return branches
 
     def find_nodes(self, query):
+        # TODO: once we can query influx better for retention periods, honor the start/end time in the FindQuery object
         with statsd.timer('service=graphite-api.action=yield_nodes.target_type=gauge.unit=ms.what=query_duration'):
             for (name, res) in self.get_leaves(query):
-                yield InfluxLeafNode(name, InfluxdbReader(self.client, name, res, self.cache, self.cheat_times))
+                yield InfluxLeafNode(name, InfluxdbReader(self.client, name, res, self.cache))
             for name in self.get_branches(query):
                 yield BranchNode(name)
 
     def fetch_multi(self, nodes, start_time, end_time):
         series = ', '.join(['"%s"' % node.path for node in nodes])
-        step = 60  # TODO: this is not ideal in all cases. for one thing, don't hardcode, for another.. how to deal with multiple steps?
+        # use the step of the node that is the most coarse
+        # not sure if there's a batter way? can we combine series with different steps (and use the optimal step for each?)
+        # probably not
+        step = max([node.reader.step for node in nodes])
         query = 'select time, value from %s where time > %ds and time < %ds order asc' % (
                 series, start_time, end_time + 1)
         logger.debug(caller='fetch_multi', query=query)
